/**
 * @file assimilator.cpp
 * @authors Martin Holkovic
 * @authors Lukas Zobal
 * @date 4. 2. 2017
 */

// This file is part of BOINC.
// http://boinc.berkeley.edu
// Copyright (C) 2015 University of California
//
// BOINC is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// BOINC is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with BOINC.  If not, see <http://www.gnu.org/licenses/>.

// A sample assimilator that:
// 1) if success, copy the output file(s) to a directory
// 2) if failure, append a message to an error log
//
// This is a sample BOINC work assimilator file edited by
// Martin Holkovic and Lukas Zobal


/**********************************/
/** @section Includes/definitions */
/**********************************/

#include <vector>
#include <string>
#include <cstdlib>
#include <math.h>
#include <boost/lexical_cast.hpp>
#include <iostream>
#include <stdint.h>
#include <inttypes.h>

#include "backend_lib.h"
#include "boinc_db.h"
#include "error_numbers.h"
#include "filesys.h"
#include "sched_msgs.h"
#include "validate_util.h"
#include "sched_config.h"
#include "sched_util.h"
#include "assimilate_handler.h"


#define SQL_BUF_SIZE 4096


/*******************************/
/** @section 3-hde table names */
/*******************************/

std::string mysql_table_job =           "fc_job";
std::string mysql_table_package =       "fc_package";
std::string mysql_table_host =          "fc_host";
std::string mysql_table_host_activity = "fc_host_activtity";
std::string mysql_table_mask =          "fc_mask";
std::string mysql_table_settings =      "fc_settings";
std::string mysql_table_hashcash =      "fc_hashcache";


/**
 * @brief PackageStatus enum
 */
enum PackageStatus
{
    Package_ready       =  0,   // Package is not running
    Package_finished    =  1,   // Password was found
    Package_exhausted   =  2,   // Keyspace was exhausted, pass not found
    Package_malformed   =  3,   // Package has incorrect input
    Package_timeout     =  4,   // time_end of the package was reached, cracking stopped
    Package_running     = 10,   // Package is running
    Package_validating  = 11,   // not used
    Package_finishing   = 12    // Packge is running, but all jobs has been generated, ending soon
};


/**
 * @brief HostStatus enum
 */
enum HostStatus
{
    Host_benchmark  = 0,
    Host_normal     = 1,
    Host_validation = 2,
    Host_done       = 3,
    Host_error      = 4
};


/*******************************/
/** @section Class definitions */
/*******************************/

/**
 * @brief MysqlJob class definiton
 */
class MysqlJob
{
    public:
        /** Constructor */
        MysqlJob(MYSQL_ROW row);

        /** Table Attributes */
        uint64_t m_id;
        uint64_t m_package_id;
        uint64_t m_workunit_id;
        uint64_t m_host_id;
        uint64_t m_boinc_host_id;

        uint64_t  m_start_index;
        uint64_t  m_hc_keyspace;
        bool      m_duplicated;
        uint64_t  m_duplicate;
        bool      m_retry;
        bool      m_finished;
};


/**
 * @brief MysqlJob constructor
 * @param row
 */
MysqlJob::MysqlJob(MYSQL_ROW row)
{
    this->m_id = boost::lexical_cast<uint64_t>(row[0]);
    this->m_package_id = boost::lexical_cast<uint64_t>(row[1]);
    this->m_workunit_id = boost::lexical_cast<uint64_t>(row[2]);
    this->m_host_id = boost::lexical_cast<uint64_t>(row[3]);
    this->m_boinc_host_id = boost::lexical_cast<uint64_t>(row[4]);
    this->m_start_index = boost::lexical_cast<uint64_t>(row[5]);
    this->m_hc_keyspace = boost::lexical_cast<uint64_t>(row[6]);
    this->m_duplicated = (row[7][0] == '0') ? true: false;
    this->m_duplicate = boost::lexical_cast<uint64_t>(row[8]);
    this->m_retry = (row[11][0] == '0') ? true: false;
    this->m_finished = (row[12][0] == '0') ? true: false;
}


/*****************************************/
/** @section Global function definitions */
/*****************************************/

void update_mysql(char * query)
{
    log_messages.printf(MSG_DEBUG, "update_mysql: %s\n", query);

    int retval = boinc_db.do_query(query);
    if (retval)
    {
        printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }
}


uint64_t get_num_from_mysql(char * query)
{
    std::cerr << __LINE__ << " - Getting number from DB: " << query << std::endl;

    uint64_t result = 0;

    //////////////////////////////////////////////////////////////////////////////////////////////////////////////
    update_mysql(query);

    MYSQL_RES* rp;
    rp = mysql_store_result(boinc_db.mysql);
    if (!rp)
    {
        printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }

    MYSQL_ROW row;
    while ((row = mysql_fetch_row(rp)))
    {
        if(row[0] != NULL)
            result = boost::lexical_cast<uint64_t>(row[0]);
    }

    mysql_free_result(rp);
    //////////////////////////////////////////////////////////////////////////////////////////////////////////////

    std::cerr << __LINE__ << " - Result: " << result << std::endl;

    return result;
}

std::string get_str_from_mysql(char * query)
{
    std::cerr << __LINE__ << " - Getting string from DB: " << query << std::endl;

    std::string result = "";

    //////////////////////////////////////////////////////////////////////////////////////////////////////////////
    update_mysql(query);

    MYSQL_RES* rp;
    rp = mysql_store_result(boinc_db.mysql);
    if (!rp)
    {
        printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }

    MYSQL_ROW row;
    while ((row = mysql_fetch_row(rp)))
    {
        if(row[0] != NULL)
            result = row[0];
    }

    mysql_free_result(rp);
    //////////////////////////////////////////////////////////////////////////////////////////////////////////////

    std::cerr << __LINE__ << " - Result: " << result << std::endl;

    return result;
}

unsigned int hc_row_count(std::string filter)
{
    char buf[SQL_BUF_SIZE];

    std::snprintf(buf, SQL_BUF_SIZE, "SELECT COUNT(*) FROM `%s` %s ;", mysql_table_hashcash.c_str(), filter.c_str());

    update_mysql(buf);

    MYSQL_RES* rp;
    rp = mysql_store_result(boinc_db.mysql);
    if (!rp)
    {
        std::printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }

    MYSQL_ROW row;

    int result = 0;
    while ((row = mysql_fetch_row(rp)))
    {
        result = boost::lexical_cast<unsigned int>(row[0]);
    }

    mysql_free_result(rp);

    return result;
}


int write_error(char* p)
{
    static FILE* f = 0;
    if (!f)
    {
        f = fopen(config.project_path("sample_results/errors"), "a");
        if (!f)
            return ERR_FOPEN;
    }

    std::fprintf(f, "%s", p);
    fflush(f);

    return 0;
}


int assimilate_handler_init(int, char**)
{
    return 0;
}


void assimilate_handler_usage()
{
    // describe the project specific arguments here
    //fprintf(stderr,
    //    "    Custom options:\n"
    //    "    [--project_option X]  a project specific option\n"
    //);
}

vector<MysqlJob *> find_jobs2(uint64_t package_id, std::string query)
{
    log_messages.printf(MSG_DEBUG, "find_jobs2: package_id-%" PRIu64 ", query-%s\n", package_id, query.c_str());

    vector<MysqlJob *> result;

    if(query != "")
    {
        query += " AND";
    }

    char buf[SQL_BUF_SIZE];

    std::snprintf(buf, SQL_BUF_SIZE, "SELECT * FROM `%s` WHERE %s host_id IN (SELECT id FROM `%s` WHERE package_id= %" PRIu64 ") ORDER BY id DESC ;", mysql_table_job.c_str(), query.c_str(), mysql_table_host.c_str(), package_id);
    update_mysql(buf);

    MYSQL_RES* rp;
    rp = mysql_store_result(boinc_db.mysql);
    if (!rp)
    {
        printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }

    MYSQL_ROW row;
    while ((row = mysql_fetch_row(rp)))
    {
        result.push_back(new MysqlJob(row));
    }

    mysql_free_result(rp);

    log_messages.printf(MSG_DEBUG, "find_jobs2-result: %lu jobs\n", result.size());

    return result;
}


// TODO: NOT USBALE
vector<MysqlJob *> find_job_duplicates2(uint64_t package_id, uint64_t host_id)
{
    log_messages.printf(MSG_DEBUG, "find_job_duplicates2: package_id-%" PRIu64 ", host_id-%" PRIu64 "\n", package_id, host_id);

    vector<MysqlJob *> result, temp;
    char buf[SQL_BUF_SIZE];

    std::snprintf(buf, SQL_BUF_SIZE, "SELECT * FROM `%s` WHERE id = %" PRIu64 " LIMIT 1;", mysql_table_job.c_str(), host_id);
    log_messages.printf(MSG_DEBUG, "find_job_duplicates2-query: %s\n", buf);

    update_mysql(buf);

    MYSQL_RES* rp;
    rp = mysql_store_result(boinc_db.mysql);
    if (!rp)
    {
        printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }

    MYSQL_ROW row;
    while ((row = mysql_fetch_row(rp)))
    {
        result.push_back(new MysqlJob(row));
    }

    mysql_free_result(rp);

    log_messages.printf(MSG_DEBUG, "find_job_duplicates2-result init: %lu jobs\n", result.size());

    for (unsigned long int index = 0; index < result.size(); ++index)
    {
        log_messages.printf(MSG_DEBUG, "find_job_duplicates2-duplicate1: %lu : %lu jobs\n", index, result.size());

        std::snprintf(buf, SQL_BUF_SIZE, "`duplicate`='%" PRIu64 "'", result[index]->m_id);
        temp = find_jobs2(package_id, buf);

        result.insert(result.end(), temp.begin(), temp.end());

        log_messages.printf(MSG_DEBUG, "find_job_duplicates2-duplicate2: %lu : %lu jobs\n", index, result.size());
    }

    return result;
}


void cancel_jobs2(vector<MysqlJob *> jobs)
{
    log_messages.printf(MSG_DEBUG, "cancel_jobs2: %lu jobs\n", jobs.size());

    std::vector<MysqlJob *>::iterator job;

    for(job = jobs.begin(); job != jobs.end(); ++job)
    {
        log_messages.printf(MSG_DEBUG, "cancel_jobs: %" PRIu64 "\n", (*job)->m_workunit_id);
        //std::cerr << __LINE__ << " cancel job-" << (*job)->m_workunit_id << std::endl;

        cancel_jobs((*job)->m_workunit_id, (*job)->m_workunit_id);
    }
}


void delete_jobs2(vector<MysqlJob *> jobs)
{
    log_messages.printf(MSG_DEBUG, "delete_jobs2: jobs:%lu\n", jobs.size());

    std::vector<MysqlJob *>::iterator job;
    char buf[SQL_BUF_SIZE];

    for(job = jobs.begin(); job != jobs.end(); ++job)
    {
        std::snprintf(buf, SQL_BUF_SIZE, "DELETE FROM `%s` WHERE id = %" PRIu64 " LIMIT 1;", mysql_table_job.c_str(), (*job)->m_id);
        update_mysql(buf);
    }
}


/**
 * @brief Sets 'finish' column to number of 'fc_job' entries
 * @param jobs Vector of 'fc_job' entries
 */
void finish_jobs(vector<MysqlJob *> jobs)
{
    log_messages.printf(MSG_DEBUG, "finish_jobs: jobs:%lu\n", jobs.size());

    std::vector<MysqlJob *>::iterator job;
    char buf[SQL_BUF_SIZE];

    for(job = jobs.begin(); job != jobs.end(); ++job)
    {
        std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET finished = 1 WHERE id = %" PRIu64 " LIMIT 1;", mysql_table_job.c_str(), (*job)->m_id);
        update_mysql(buf);
    }
}


/**
 * @brief Set 'retry' column to one job entry
 * @param workunit_id the BOINC workunit ID used for job identification
 */
void set_retry_job(uint64_t workunit_id)
{
    char buf[SQL_BUF_SIZE];
    std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET retry = 1 WHERE workunit_id = %" PRIu64 " LIMIT 1;", mysql_table_job.c_str(), workunit_id);
    update_mysql(buf);
}


/**
 *
 *
 */
void plan_new_benchamrk(uint64_t host_id)
{
    char buf[SQL_BUF_SIZE];

    /** Plan another benchmark in future */
    std::snprintf(buf, SQL_BUF_SIZE, "SELECT power FROM `%s` WHERE id = %" PRIu64 " LIMIT 1;", mysql_table_host.c_str(), host_id);
    uint64_t power = get_num_from_mysql(buf);
    uint64_t minutes = pow(2, power);
    ++power;

    std::cerr << __LINE__ << " - Planning next benchmark in " << minutes << " minutes" << std::endl;

    std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET power = %" PRIu64 ", time = now() + INTERVAL %" PRIu64 " MINUTE WHERE id = %" PRIu64 ";", mysql_table_host.c_str(), power, minutes, host_id);
    update_mysql(buf);
}


void update_power(uint64_t host_id, uint64_t workId, double elapsed_time)
{
    uint64_t power;
    //uint64_t actualPower;
    uint64_t count;
    char buf[SQL_BUF_SIZE];
    /*
    std::snprintf(buf, SQL_BUF_SIZE, "SELECT `power` FROM `fc_host` WHERE id = %" PRIu64 " LIMIT 1;", host_id);
    power = get_num_from_mysql(buf);
    */
    std::snprintf(buf, SQL_BUF_SIZE, "SELECT hc_keyspace FROM `%s` WHERE workunit_id = %" PRIu64 " LIMIT 1;", mysql_table_job.c_str(), workId);
    count = get_num_from_mysql(buf);

    power = count / elapsed_time;

    std::cerr << __LINE__ << " - Updating power: " << power << " to host_id: " << host_id << std::endl;

    std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET power = %" PRIu64 ", time = now() WHERE id = %" PRIu64 " LIMIT 1;", mysql_table_host.c_str(), power, host_id);
    update_mysql(buf);
}


/********************************/
/** @section Assimilate Handler */
/********************************/

int assimilate_handler(WORKUNIT& wu, vector<RESULT>& /*results*/, RESULT& canonical_result)
{
    std::cerr << std::endl;
    std::cerr << __LINE__ << " - Incoming workunit result with id " << wu.id << std::endl;

    int retval;         // fscanf return values
    char buf[SQL_BUF_SIZE];     // SQL queries
    char mode;          // Result mode (b/n)
    int code;           // Result status code (0 OK, 1 notFound/err, etc.)
    unsigned int power;         // Host power (bench)
    double cracking_time;       // Cracking time
    char param[256];    // Other string params (err msg)

    uint64_t host_id;
    uint64_t package_id;
    std::vector<MysqlJob *> jobs;

    retval = boinc_mkdir(config.project_path("sample_results"));

    if (retval)
    {
        std::cerr << __LINE__ << " - Failed to create result dir" << std::endl;
        return retval;
    }

    if (!wu.canonical_resultid)
    {
        std::snprintf(buf, SQL_BUF_SIZE, "%s: 0x%x\n", wu.name, wu.error_mask);

        std::cerr << __LINE__ << " - Result error: " << buf;
        std::cerr << std::flush;

        return write_error(buf);
    }

    vector<OUTPUT_FILE_INFO> output_files;
    get_output_file_infos(canonical_result, output_files);
    unsigned long int n = output_files.size();

    /** Find out who sent the result */
    std::snprintf(buf, SQL_BUF_SIZE, "SELECT host_id FROM `%s` WHERE workunit_id = %lu LIMIT 1;", mysql_table_job.c_str(), wu.id);
    host_id = get_num_from_mysql(buf);

    std::cerr << __LINE__ << " - From host " << host_id << std::endl;

    /** Find out which package is the result from */
    std::snprintf(buf, SQL_BUF_SIZE, "SELECT package_id FROM `%s` WHERE workunit_id = %lu LIMIT 1;", mysql_table_job.c_str(), wu.id);
    package_id = get_num_from_mysql(buf);

    std::cerr << __LINE__ << " - From package " << package_id << std::endl;

    if(!package_id)
    {
        log_messages.printf(MSG_CRITICAL, "not possible to find package for workunit_id = %lu\n", wu.id);

        std::cerr << __LINE__ << " - Error: unable to find such workunit_id - " << wu.id << std::endl;

        return 0;
    }

    // Read the fc_settings
    std::snprintf(buf, SQL_BUF_SIZE, "SELECT delete_finished_jobs FROM `%s` LIMIT 1;", mysql_table_settings.c_str());
    uint64_t deleteFlag = get_num_from_mysql(buf);

    for (unsigned int i = 0; i < n; ++i)
    {
        OUTPUT_FILE_INFO& fi = output_files[i];

        /** Open the output file */
        FILE* f = std::fopen(fi.path.c_str(), "r");

        if(f == NULL)
        {
            std::cerr << __LINE__ << " - Error: unable to open output file, code " << errno << std::endl;
            return 0;
        }

        std::cerr << __LINE__ << " - Output file opened: " << fi.path.c_str() << std::endl;

        /** Read the mode */
        retval = std::fscanf(f,"%c\n", &mode);
        if (retval != 1)
        {
            std::cerr << __LINE__ << " - ERROR: Failed to read output mode." << std::endl;
            set_retry_job(wu.id);
            std::fclose(f);
            break;
        }

        /** BENCHMARK mode */
        if(mode == 'b')
        {
            std::cerr << __LINE__ << " - WU_MODE: BENCHMARK (b)" << std::endl;
            log_messages.printf(MSG_DEBUG, "mode b\n");

            /** Read the code */
            retval = std::fscanf(f,"%d\n", &code);
            if (retval != 1)
            {
                std::cerr << __LINE__ << " - ERROR: Failed to read status code." << std::endl;
                plan_new_benchamrk(host_id);
                break;
            }

            /** Result OK */
            if(code == 0)
            {
                std::cerr << __LINE__ << " - B_CODE: 0 (bench OK)" << std::endl;
                log_messages.printf(MSG_DEBUG, "code %d\n", code);

                /** Read Power (p/s) */
                retval = std::fscanf(f,"%u\n", &power);
                if (retval != 1)
                {
                    std::cerr << __LINE__ << " - ERROR: Failed to read host power." << std::endl;
                    plan_new_benchamrk(host_id);
                    break;
                }

                std::cerr << __LINE__ << " - New host power: " << power << std::endl;

                /** In case of mask attack, convert power to mask indices/s */
                std::snprintf(buf, SQL_BUF_SIZE, "SELECT attack_mode FROM `%s` WHERE id = %" PRIu64 " LIMIT 1;", mysql_table_package.c_str(), package_id);
                uint64_t attack_mode = get_num_from_mysql(buf);

                if (attack_mode == 3)
                {
                    std::snprintf(buf, SQL_BUF_SIZE, "SELECT keyspace FROM `%s` WHERE id = %" PRIu64 " LIMIT 1;", mysql_table_package.c_str(), package_id);
                    uint64_t keyspace = get_num_from_mysql(buf);

                    std::snprintf(buf, SQL_BUF_SIZE, "SELECT hc_keyspace FROM `%s` WHERE id = %" PRIu64 " LIMIT 1;", mysql_table_package.c_str(), package_id);
                    uint64_t hc_keyspace = get_num_from_mysql(buf);

                    uint64_t factor = keyspace / hc_keyspace;

                    if (factor > 0)
                    {
                        power /= factor;
                        std::cerr << __LINE__ << " - Updating power for mask-attack with ratio: 1/" << factor << ", new power: " << power << std::endl;
                    }
                }

                std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET power = %u, status = %d, time = now() WHERE id = %" PRIu64 " ;", mysql_table_host.c_str(), power, Host_normal, host_id);
                update_mysql(buf);

                /** Read cracking_time */
                retval = std::fscanf(f,"%lf\n", &cracking_time);
                if (retval != 1)
                {
                    std::cerr << __LINE__ << " - ERROR: Failed to read cracking_time." << std::endl;
                    plan_new_benchamrk(host_id);
                    break;
                }
                std::cerr << __LINE__ << " - Benchmark time: " << cracking_time << "s" << std::endl;

                std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET cracking_time = cracking_time + %lf WHERE id = %" PRIu64 " ;", mysql_table_package.c_str(), cracking_time, package_id);
                update_mysql(buf);

                std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET cracking_time = %lf WHERE `workunit_id` = %lu LIMIT 1;", mysql_table_job.c_str(), cracking_time, wu.id);
                update_mysql(buf);
            }

            /** Result ERROR */
            else
            {
                std::cerr << __LINE__ << " - B_CODE: " << code <<" (bench ERROR)" << std::endl;
                log_messages.printf(MSG_DEBUG, "code %d\n", code);

                plan_new_benchamrk(host_id);
            }
        }

        /** NORMAL mode */
        else if(mode == 'n')
        {
            std::cerr << __LINE__ << " - WU_MODE: NORMAL (n)" << std::endl;
            log_messages.printf(MSG_DEBUG, "mode n\n");

            /** Read the code */
            retval = std::fscanf(f,"%d\n", &code);
            if (retval != 1)
            {
                std::cerr << __LINE__ << " - ERROR: Failed to read status code." << std::endl;
                std::fclose(f);
                set_retry_job(wu.id);
                return 0;
            }

            /** Password FOUND */
            if(code == 0)
            {
                std::cerr << __LINE__ << " - N_CODE: 0 (pass FOUND)" << std::endl;
                log_messages.printf(MSG_DEBUG, "code %d\n", code);

                /** Read the result (password) */
                retval = std::fscanf(f,"%255s\n", param);
                if (retval != 1)
                {
                    std::cerr << __LINE__ << " - ERROR: Failed to read password." << std::endl;
                    std::fclose(f);
                    set_retry_job(wu.id);
                    return 0;
                }

                std::cerr << __LINE__ << " - Encoded password: " << param << std::endl;
                std::cerr << __LINE__ << " - Setting package " << package_id << " to Finished status" << std::endl;

                /** Without validation */
                std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET status = %d, result = '%s', time_end = now() WHERE id = %" PRIu64 " ;", mysql_table_package.c_str(), Package_finished, param, package_id);
                update_mysql(buf);

                std::cerr << __LINE__ << " - Canceling all jobs for package_id " << package_id << std::endl;
                jobs = find_jobs2(package_id, "");
                cancel_jobs2(jobs);

                if (deleteFlag)
                    delete_jobs2(jobs);
                else
                    finish_jobs(jobs);

                /** Read cracking_time */
                retval = std::fscanf(f,"%lf\n", &cracking_time);
                if (retval == 1)
                {
                    std::cerr << __LINE__ << " - Cracking time: " << cracking_time << "s" << std::endl;

                    std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET cracking_time = cracking_time + %lf WHERE id = %" PRIu64 " ;", mysql_table_package.c_str(), cracking_time, package_id);
                    update_mysql(buf);

                    std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET cracking_time = %lf WHERE `workunit_id` = %lu LIMIT 1;", mysql_table_job.c_str(), cracking_time, wu.id);
                    update_mysql(buf);
                }
                else
                    std::cerr << __LINE__ << " - WARNING: cracking time of successful job could not be read" << std::endl;

                /** Insert the result into hashcache table */
                std::snprintf(buf, SQL_BUF_SIZE, "SELECT hash_type FROM `%s` WHERE id = %" PRIu64 " LIMIT 1;", mysql_table_package.c_str(), package_id);
                uint64_t hash_type = get_num_from_mysql(buf);

                std::snprintf(buf, SQL_BUF_SIZE, "SELECT hash FROM `%s` WHERE id = %" PRIu64 " LIMIT 1;", mysql_table_package.c_str(), package_id);
                std::string hash = get_str_from_mysql(buf);

                std::snprintf(buf, SQL_BUF_SIZE, "WHERE hash_type = '%" PRIu64 "' AND hash = '%s' AND result = '%s'", hash_type, hash.c_str(), param);
                if(hc_row_count(buf) == 0)
                {
                    std::cerr << __LINE__ << " - Adding result to hashcache." << std::endl;
                    std::snprintf(buf, SQL_BUF_SIZE, "INSERT INTO `%s` (`hash_type`,`hash`,`result`) VALUES (%" PRIu64 ", '%s','%s');",
                                mysql_table_hashcash.c_str(), hash_type, hash.c_str(), param);
                    update_mysql(buf);
                }
                else
                {
                    std::cerr << __LINE__ << " - Result already in hashcache." << std::endl;
                }

                /** Validation */
/*
                std::sprintf(buf, "UPDATE `fc_package` SET `status`='%d',`result`='%s' WHERE id = %" PRIu64 " ;", Package_validating, param, package_id);
                update_mysql(buf);

                retval = fscanf(f,"%255s\n", param);
                std::sprintf(buf, "UPDATE `fc_package` SET `cracking_time`=`cracking_time`+%s WHERE `id` = %" PRIu64 " ;", param, package_id);
                update_mysql(buf);
                std::sprintf(buf, "UPDATE `fc_workunit` SET `fitTime`=%s WHERE `workunit_id` = %lu ;", param, wu.id);
                std::cerr << __LINE__ << " " << buf << std::endl;
                update_mysql(buf);

                // vyhledat najnovsi ulohu s inym host_id
                uint64_t anotherHostId;
                std::sprintf(buf, "SELECT `host_id` FROM `fc_job` WHERE host_id != %" PRIu64 " AND host_id IN (SELECT id FROM `fc_host` WHERE `package_id`='%" PRIu64 "') ORDER BY id DESC LIMIT 1;", host_id, package_id);
                anotherHostId = get_num_from_mysql(buf);


                if(anotherHostId != 0) {
                    std::sprintf(buf, "`host_id`='%" PRIu64 "'", anotherHostId);
                    jobs = find_jobs2(package_id, buf);
                    std::cerr << __LINE__ << std::endl;
                    cancel_jobs2(jobs);
                    add_jobs_to_retry2(package_id, jobs);
                    delete_jobs2(jobs);
                }

                update_power(host_id, wu.id, canonical_result.elapsed_time);
*/
            }

            /** Password NOT yet found */
            else if(code == 1)
            {
                std::cerr << __LINE__ << " - N_CODE: 1 (pass NOT found)" << std::endl;
                log_messages.printf(MSG_DEBUG, "code %d\n", code);

                /** Read cracking_time */
                retval = std::fscanf(f,"%lf\n", &cracking_time);
                if (retval == 1)
                {
                    std::cerr << __LINE__ << " - Cracking time: " << cracking_time << "s" << std::endl;

                    std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET cracking_time = cracking_time + %lf WHERE id = %" PRIu64 " ;", mysql_table_package.c_str(), cracking_time, package_id);
                    update_mysql(buf);

                    std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET cracking_time = %lf WHERE `workunit_id` = %lu LIMIT 1;", mysql_table_job.c_str(), cracking_time, wu.id);
                    update_mysql(buf);

                    /** Update the host power */
                    update_power(host_id, wu.id, canonical_result.elapsed_time);
                }
                else
                    std::cerr << __LINE__ << " - WARNING: Was NOT able to read cracking_time of a job, not updating power" << std::endl;

                // Update package progress
                std::snprintf(buf, SQL_BUF_SIZE, "SELECT hc_keyspace FROM `%s` WHERE workunit_id = %lu LIMIT 1;", mysql_table_job.c_str(), wu.id);
                uint64_t hc_keyspace = get_num_from_mysql(buf);

                std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET indexes_verified = indexes_verified + %" PRIu64 " WHERE id = %" PRIu64 " LIMIT 1;", mysql_table_package.c_str(), hc_keyspace, package_id);
                update_mysql(buf);

            }

            // TODO >>> ADD MALFORMED PACKAGE STATE PARSING

            /** Host error */
            else
            {
                /** Cancel and delete host jobs and put them to Retry */
                log_messages.printf(MSG_DEBUG, "code %d\n", code);
                std::cerr << __LINE__ << " - N_CODE: " << code << " (host ERROR)" << std::endl;

                std::snprintf(buf, SQL_BUF_SIZE, "`host_id`='%" PRIu64 "'", host_id);

                std::cerr << __LINE__ << " - Canceling all jobs for host_id " << host_id << " with package_id" << package_id << std::endl;
                jobs = find_jobs2(package_id, buf);
                cancel_jobs2(jobs);

                std::cerr << __LINE__ << " - Adding them to Retry and deleting them from Jobs" << std::endl;
                std::vector<MysqlJob *>::iterator jobIt;
                for(jobIt = jobs.begin(); jobIt != jobs.end(); ++jobIt)
                    set_retry_job((*jobIt)->m_workunit_id);

                std::cerr << __LINE__ << " - Updating power to 0 and status to Benchmark for host_id " << host_id << std::endl;
                std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET power = 0, status = %d WHERE id = %" PRIu64 " ;", mysql_table_host.c_str(), Host_benchmark, host_id);
                update_mysql(buf);

                // Do not set pacakge to finished/delete
                std::fclose(f);
                return 0;
            }
        }

        /** VALIDATION mode */
        /** NOT USED */
        else if(mode == 'v')
        {
            std::cerr << __LINE__ << " - WU_MODE: VALIDATION (v)" << std::endl;
            log_messages.printf(MSG_DEBUG, "mode v\n");

            /** Read the code */
            retval = fscanf(f,"%d\n", &code);

            /** Password is Validated */
            if(code == 0)
            {
                std::cerr << __LINE__ << " - V_CODE: 0 (pass VALIDATED)" << std::endl;
                log_messages.printf(MSG_DEBUG, "code %d\n", code);

                /** Read the Password */
                retval = fscanf(f,"%255s\n", param);
                std::cerr << __LINE__ << " - Encoded password: " << param << std::endl;
                std::cerr << __LINE__ << " - Setting package " << package_id << " to Finished status" << std::endl;

                std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `fc_package` SET `status`='%d',`result`='%s',`time_end`=now() WHERE id = %" PRIu64 " ;", Package_finished, param, package_id);
                update_mysql(buf);

                std::cerr << __LINE__ << " - Canceling all jobs for package_id " << package_id << std::endl;
                jobs = find_jobs2(package_id, "");
                cancel_jobs2(jobs);
                delete_jobs2(jobs);

                // TODO: useless delete?

                //
                // TODO:  MAZAT V ZÁVISLOSTI NA NASTAVENÍ delete_finished_jobs ve fc_settings!
                //

                std::cerr << __LINE__ << " - Deleting all jobs for package_id " << package_id << std::endl;
                std::snprintf(buf, SQL_BUF_SIZE, "DELETE FROM `fc_job` WHERE host_id IN (SELECT id FROM `fc_host` WHERE `package_id`='%" PRIu64 "');", package_id);
                update_mysql(buf);
/*
                std::cerr << __LINE__ << " - Deleting all retries for package_id " << package_id << std::endl;
                std::sprintf(buf, "DELETE FROM `fc_retry` WHERE host_id IN (SELECT id FROM `fc_host` WHERE `package_id`='%" PRIu64 "');", package_id);
                update_mysql(buf);
*/
                std::cerr << __LINE__ << " - Deleting all hosts for package_id " << package_id << std::endl;
                std::snprintf(buf, SQL_BUF_SIZE, "DELETE FROM `fc_host` WHERE `package_id`='%" PRIu64 "';", package_id);
                update_mysql(buf);
/*
                std::cerr << __LINE__ << " - TODO?: Set another (running) package to run in 1 minute" << std::endl;
                std::sprintf(buf, "UPDATE `fc_package` SET `time_start`=now() + INTERVAL 1 MINUTE WHERE status = '%d' ;", Package_running);
                update_mysql(buf);
*/
                /** Read FitTime */
                retval = fscanf(f,"%255s\n", param);
                std::cerr << __LINE__ << " - Validating time: " << param << "s" << std::endl;

                std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `fc_package` SET `cracking_time`=`cracking_time`+%s WHERE `id` = %" PRIu64 " ;", param, package_id);
                update_mysql(buf);

                std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `fc_workunit` SET `fitTime`=%s WHERE `workunit_id` = %lu ;", param, wu.id); // TODO: Asi odstranit, neni to to same jako cracking_time?
                update_mysql(buf);
            }

            /** Password is invalidated */
            else if(code == 1)
            {
                std::cerr << __LINE__ << " - V_CODE: 1 (pass INVALIDATED)" << std::endl;
                log_messages.printf(MSG_DEBUG, "code %d\n", code);

                std::cerr << __LINE__ << " - Setting package " << package_id << " to Running status" << std::endl;
                std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `fc_package` SET `status`='%d' WHERE id = %" PRIu64 " ;", Package_running, package_id);
                update_mysql(buf);

                std::cerr << __LINE__ << " - Canceling all validating jobs for package_id " << package_id << std::endl;
                jobs = find_jobs2(package_id, "`from`='1' AND `count`='0'");
                cancel_jobs2(jobs);
                delete_jobs2(jobs);

                retval = fscanf(f,"%255s\n", param);
                std::cerr << __LINE__ << " - Validating time: " << param << "s" << std::endl;

                std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `fc_package` SET `cracking_time`=`cracking_time`+%s WHERE `id` = %" PRIu64 " ;", param, package_id);
                update_mysql(buf);

                std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `fc_workunit` SET `fitTime`=%s WHERE `workunit_id` = %lu ;", param, wu.id);
                update_mysql(buf);
            }

            /** Error occured */
            else
            {
                std::cerr << __LINE__ << " - V_CODE: " << code << " (host ERROR)" << std::endl;
                std::cerr << __LINE__ << " - Waiting for another validation" << std::endl;
            }
        }

        if (deleteFlag)
        {
            /** Delete finished job */
            std::cerr << __LINE__ << " - Deleting assimilated job with workId " << wu.id << std::endl;
            std::snprintf(buf, SQL_BUF_SIZE, "DELETE FROM `%s` WHERE workId = %lu ;", mysql_table_job.c_str(), wu.id);
            update_mysql(buf);
        }
        else
        {
            /** Set job finished flag */
            std::cerr << __LINE__ << " - Setting finished flag to assimilated job with workunit_id " << wu.id << std::endl;
            std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET finished = 1 WHERE workunit_id = %lu ;", mysql_table_job.c_str(), wu.id);
            update_mysql(buf);
        }

        std::fclose(f);
    }

    return 0;
}
