/**
 * @file sample_work_generator.cpp
 * @authors Lukas Zobal
 * @authors Martin Holkovic
 * @date 18. 8. 2017
 */

// This file is part of BOINC.
// http://boinc.berkeley.edu
// Copyright (C) 2008 University of California
//
// BOINC is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// BOINC is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with BOINC.  If not, see <http://www.gnu.org/licenses/>.

// work_generator: BOINC work generator.
//
// --app name               app name (default example_app)
// --in_template_file       input template file (default example_app_in)
// --out_template_file      output template file (default example_app_out)
// -d N                     log verbosity level (0..4)
// --help                   show usage
// --version                show version
//
// - Runs as a daemon, and creates an unbounded supply of work.
//   It attempts to maintain a "cushion" of 100 unsent job instances
//   for the given app.
//   (your app may not work this way; e.g. you might create work in batches)
// - Creates a new input file for each job;
//   the file (and the workunit names) contain a timestamp
//   and sequence number, so they're unique.
//
// This is a sample BOINC work generator file edited by
// Martin Holkovic and Lukas Zobal


/**********************************/
/** @section Includes/definitions */
/**********************************/

#include <boost/archive/iterators/base64_from_binary.hpp>
#include <boost/archive/iterators/binary_from_base64.hpp>
#include <boost/archive/iterators/transform_width.hpp>
#include <boost/archive/iterators/insert_linebreaks.hpp>
#include <boost/archive/iterators/remove_whitespace.hpp>
#include <boost/lexical_cast.hpp>

#include "sample_work_generator.h"

//#define RUN_SECONDS 60
//#define RETRY_RUN_SECONDS 100

#define REPLICATION_FACTOR  1   // Number of instances of each job
#define SQL_BUF_SIZE 4096


using namespace boost::archive::iterators;


/**
 * @brief PackageStatus enum
 */
enum PackageStatus
{
    Package_ready       =  0,   // Package is not running
    Package_finished    =  1,   // Password was found
    Package_exhausted   =  2,   // Keyspace was exhausted, pass not found
    Package_malformed   =  3,   // Package has incorrect input
    Package_timeout     =  4,   // time_end of the package was reached, cracking stopped
    Package_running     = 10,   // Package is running
    Package_validating  = 11,   // not used
    Package_finishing   = 12    // Packge is running, but all jobs has been generated, ending soon
};


/**
 * @brief HostStatus enum
 */
enum HostStatus
{
    Host_benchmark  = 0,
    Host_normal     = 1,
    Host_validation = 2,
    Host_done       = 3,
    Host_error      = 4
};


/*****************************************/
/** @section Global function definitions */
/*****************************************/

std::string decode(std::string base64)
{
    typedef transform_width< binary_from_base64<remove_whitespace<std::string::const_iterator> >, 8, 6 > it_binary_t;
    unsigned int paddChars = std::count(base64.begin(), base64.end(), '=');

    std::replace(base64.begin(),base64.end(),'=','A'); // replace '=' by base64 encoding of '\0'

    std::string result(it_binary_t(base64.begin()), it_binary_t(base64.end())); // decode
    result.erase(result.end()-paddChars,result.end());  // erase padding '\0' characters

    return result;
}


uint64_t get_num_from_mysql(char * query)
{
    log_messages.printf(MSG_DEBUG, "get_num_from_mysql: %s\n", query);

    uint64_t result = 0;

    //////////////////////////////////////////////////////////////////////////////////////////////////////////////
    update_mysql(query);

    MYSQL_RES* rp;
    rp = mysql_store_result(boinc_db.mysql);
    if (!rp)
    {
        std::printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }

    MYSQL_ROW row;
    while ((row = mysql_fetch_row(rp)))
    {
        if(row[0] != NULL)
            result = boost::lexical_cast<uint64_t>(row[0]);
    }

    mysql_free_result(rp);
    //////////////////////////////////////////////////////////////////////////////////////////////////////////////

    log_messages.printf(MSG_DEBUG, "get_num_from_mysql-result: %" PRIu64 "\n", result);
    return result;
}


void update_mysql(char * query)
{
    log_messages.printf(MSG_DEBUG, "update_mysql: %s\n", query);

    int retval = boinc_db.do_query(query);
    if (retval)
    {
        std::printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }
}


int make_bench(MysqlJob * job, MysqlPackage * package)
{
    std::cerr << __LINE__ << " - make_bench(): " << "package_id = " << package->m_id << ", host_id = " << job->m_boinc_host_id << std::endl;

    DB_WORKUNIT wu;
    char name1[SQL_BUF_SIZE], path[SQL_BUF_SIZE];
    const char* infiles[1];
    int retval;

    /** Make a unique name for the job and its config file */
    std::snprintf(name1, SQL_BUF_SIZE, "%s_%d_%d", app_name, start_time, seqno++);

    /** Append bech mode to the config file */
    retval = config.download_path(name1, path);
    if (retval)
        return retval;

    std::ofstream f;
    f.open(path);

    std::cerr << "CONFIG FORMAT FOLLOWS:" << std::endl;

    /** Output original config from DB */
    f << package->m_config;
    std::cerr << package->m_config;

    /** Output mode */
    f << "|||mode|String|1|b|||\n";
    std::cerr << "|||mode|String|1|b|||" << std::endl;

    f.close();

    /** Fill in the job parameters */
    wu.clear();
    wu.appid = app.id;
    safe_strcpy(wu.name, name1);
    wu.rsc_fpops_est = 1e12;
    wu.rsc_fpops_bound = 1e14;
    wu.rsc_memory_bound = 1e8;
    wu.rsc_disk_bound = 1e8;
    wu.delay_bound = 86400;
    wu.min_quorum = REPLICATION_FACTOR;
    wu.target_nresults = REPLICATION_FACTOR;
    wu.max_error_results = REPLICATION_FACTOR*4;
    wu.max_total_results = REPLICATION_FACTOR*8;
    wu.max_success_results = REPLICATION_FACTOR*4;
    infiles[0] = name1;

    /** Register the job with BOINC */
    std::snprintf(path, SQL_BUF_SIZE, "templates/%s", out_template_file);
    retval = create_work(
        wu,
        in_template_bench,
        path,
        config.project_path(path),
        infiles,
        1,
        config
    );

    restrict_wu_to_host(wu, job->m_boinc_host_id);

    job->m_workunit_id = wu.id;
    job->add_new_job(package->m_id);

    return 0;
}


int make_job(MysqlJob * job, MysqlPackage * package)
{
    switch (package->m_attack_mode)
    {
        // Dictionary attack
        case 0:
            make_job_dict(job, package);
            break;

        // Mask attack
        case 3:
            make_job_mask(job, package);
            break;

        default:
            std::cerr << "ERROR: Such attack: " << package->m_attack << " not supported!" << std::endl;
            package->update_status_running_package(Package_malformed);
            break;
    }

    return 0;
}


int make_job_dict(MysqlJob * job, MysqlPackage * package)
{
    std::cerr << __LINE__ << " - make_job_dict(): " << "package_id = " << package->m_id << ", host_id = " << job->m_boinc_host_id << std::endl;

    DB_WORKUNIT wu;
    char name1[SQL_BUF_SIZE],name2[SQL_BUF_SIZE],name3[SQL_BUF_SIZE], path[SQL_BUF_SIZE];
    const char* infiles[3];
    int retval;

    /** Make a unique name for the job and its input file */
    std::snprintf(name1, SQL_BUF_SIZE, "%s_%d_%d", app_name, start_time, seqno++);
    std::snprintf(name2, SQL_BUF_SIZE, "%s_%d_%d", app_name, start_time, seqno++);
    std::snprintf(name3, SQL_BUF_SIZE, "%s_%d_%d", app_name, start_time, seqno++);

    /** Append mode to config */
    retval = config.download_path(name1, path);
    if (retval)
        return retval;

    std::ofstream f;
    f.open(path);

    std::cerr << "CONFIG FORMAT FOLLOWS:" << std::endl;

    /** Output original config from DB */
    f << package->m_config;
    std::cerr << package->m_config;

    /** Output mode */
    f << "|||mode|String|1|n|||\n";
    std::cerr << "|||mode|String|1|n|||" << std::endl;

    f.close();

    /** Create data file */
    retval = config.download_path(name2, path);
    if (retval)
        return retval;

    f.open(path);

    if(package->m_hash != "")
        f << package->m_hash;

    f.close();

    /** Create dictionary file. */
    retval = config.download_path(name3, path);
    if (retval)
        return retval;

    f.open(path);

    if(package->m_dict1 != "")
    {
        std::cerr << __LINE__ << " - Creating dictionary fragment" << std::endl;
        std::ifstream dictFile;

        std::cerr << __LINE__ << " - Opening dictionary " << (DICT_DIR + package->m_dict1) << std::endl;
        dictFile.open((DICT_DIR + package->m_dict1).c_str());

        if (dictFile.is_open())
        {
            std::cerr << __LINE__ << " - " << package->m_dict1 << " dictionary opened" << std::endl;

            // Ignore 'start_index' passwords
            uint64_t currentIndex;

            std::cerr << __LINE__ << " - Skipping " << job->m_start_index << " passwords" << std::endl;
            std::string passwd;
            for (currentIndex = 0; currentIndex < job->m_start_index; ++currentIndex)
            {
                std::getline(dictFile, passwd);
            }

            // No more passwords in dictionary, set package to Finishing
            if (dictFile.eof())
            {
                std::cerr << __LINE__ << " - START_INDEX parameter is too far away, setting package to finishing" << std::endl;
                package->update_status_running_package(Package_finishing);
                return 0;
            }

            std::cerr << __LINE__ << " - Adding " << job->m_hc_keyspace << " passwords to host dict file" << std::endl;

            // Add 'keyspace' passwords to dict file
            for (currentIndex = 0; currentIndex < job->m_hc_keyspace; ++currentIndex)
            {
                std::getline(dictFile, passwd);
                if (passwd != "")
                    f << passwd << '\n';

                // End of dictionary, set package to finishing
                if (dictFile.eof())
                {
                    std::cerr << __LINE__ << " - Ate all passwords, setting package to finishing" << std::endl;
                    package->update_status_running_package(Package_finishing);

                    job->m_hc_keyspace = currentIndex;
                    package->update_index(package->m_hc_keyspace);

                    break;
                }
            }

            std::cerr << "Done. Closing file." << std::endl;

            dictFile.close();
        }
    }

    f.close();

    /** Fill in the job parameters */
    wu.clear();
    wu.appid = app.id;
    safe_strcpy(wu.name, name1);
    wu.rsc_fpops_est = 1e12;
    wu.rsc_fpops_bound = 1e14;
    wu.rsc_memory_bound = 1e8;
    wu.rsc_disk_bound = 1e8;
    wu.delay_bound = 86400;
    wu.min_quorum = REPLICATION_FACTOR;
    wu.target_nresults = REPLICATION_FACTOR;
    wu.max_error_results = REPLICATION_FACTOR*4;
    wu.max_total_results = REPLICATION_FACTOR*8;
    wu.max_success_results = REPLICATION_FACTOR*4;
    infiles[0] = name1;
    infiles[1] = name2;
    infiles[2] = name3;

    /** Register the job with BOINC */
    std::snprintf(path, SQL_BUF_SIZE, "templates/%s", out_template_file);
    retval = create_work(
        wu,
        in_template_dict,
        path,
        config.project_path(path),
        infiles,
        3,
        config
    );

    restrict_wu_to_host(wu, job->m_boinc_host_id);

    job->m_workunit_id = wu.id;
    job->add_new_job(package->m_id);

    return 0;
}

int make_job_mask(MysqlJob * job, MysqlPackage * package)
{
    std::cerr << __LINE__ << " - make_job_mask(): " << "package_id = " << package->m_id << ", host_id = " << job->m_boinc_host_id << std::endl;

    DB_WORKUNIT wu;
    char name1[SQL_BUF_SIZE],name2[SQL_BUF_SIZE], path[SQL_BUF_SIZE];
    const char* infiles[2];
    int retval;

    /** Make a unique name for the job and its input file */
    std::snprintf(name1, SQL_BUF_SIZE, "%s_%d_%d", app_name, start_time, seqno++);
    std::snprintf(name2, SQL_BUF_SIZE, "%s_%d_%d", app_name, start_time, seqno++);

    /** Append mode, mask, start_index, stop_index to config */
    retval = config.download_path(name1, path);
    if (retval)
        return retval;

    std::ofstream f;
    f.open(path);

    std::cerr << "CONFIG FORMAT FOLLOWS:" << std::endl;

    /** Output original config from DB */
    f << package->m_config;
    std::cerr << package->m_config;

    /** Output mode */
    f << "|||mode|String|1|n|||\n";
    std::cerr << "|||mode|String|1|n|||" << std::endl;

    /** Output mask */
    f << "|||mask|String|" << package->m_masks[0]->m_mask.length() << "|" << package->m_masks[0]->m_mask << "|||\n";
    std::cerr << "|||mask|String|" << package->m_masks[0]->m_mask.length() << "|" << package->m_masks[0]->m_mask << "|||" << std::endl;

    /** Output start_index */
    int digits = 0;
    uint64_t num = job->m_start_index;
    do { num /= 10; ++digits; } while (num != 0);    // Count digits
    f << "|||start_index|BigUInt|" << digits << "|" << job->m_start_index << "|||\n";
    std::cerr << "|||start_index|BigUInt|" << digits << "|" << job->m_start_index << "|||" << std::endl;

    /** Output stop_index - only if it is not the last job in the current mask */
    if (package->m_masks[0]->m_current_index < package->m_masks[0]->m_hc_keyspace)
    {
        digits = 0;
        num = job->m_hc_keyspace;
        do { num /= 10; ++digits; } while (num != 0);    // Count digits
        f << "|||hc_keyspace|BigUInt|" << digits << "|" << job->m_hc_keyspace << "|||\n";
        std::cerr << "|||hc_keyspace|BigUInt|" << digits << "|" << job->m_hc_keyspace << "|||" << std::endl;
    }
    else    /** Otherwise, send whole mask_hc_keyspace for correct progress calculation, --limit is omitted */
    {
        digits = 0;
        num = package->m_masks[0]->m_hc_keyspace;
        do { num /= 10; ++digits; } while (num != 0);    // Count digits
        f << "|||mask_hc_keyspace|BigUInt|" << digits << "|" << package->m_masks[0]->m_hc_keyspace << "|||\n";
        std::cerr << "|||mask_hc_keyspace|BigUInt|" << digits << "|" << package->m_masks[0]->m_hc_keyspace << "|||" << std::endl;
    }

    f.close();

    /** Create data file */
    retval = config.download_path(name2, path);
    if (retval)
        return retval;

    f.open(path);

    if(package->m_hash != "")
        f << package->m_hash;

    f.close();


    /** Fill in the job parameters */
    wu.clear();
    wu.appid = app.id;
    safe_strcpy(wu.name, name1);
    wu.rsc_fpops_est = 1e12;
    wu.rsc_fpops_bound = 1e14;
    wu.rsc_memory_bound = 1e8;
    wu.rsc_disk_bound = 1e8;
    wu.delay_bound = 86400;
    wu.min_quorum = REPLICATION_FACTOR;
    wu.target_nresults = REPLICATION_FACTOR;
    wu.max_error_results = REPLICATION_FACTOR*4;
    wu.max_total_results = REPLICATION_FACTOR*8;
    wu.max_success_results = REPLICATION_FACTOR*4;
    infiles[0] = name1;
    infiles[1] = name2;

    /** Register the job with BOINC */
    std::snprintf(path, SQL_BUF_SIZE, "templates/%s", out_template_file);
    retval = create_work(
        wu,
        in_template_mask,
        path,
        config.project_path(path),
        infiles,
        2,
        config
    );

    restrict_wu_to_host(wu, job->m_boinc_host_id);

    job->m_workunit_id = wu.id;
    job->add_new_job(package->m_id);

    return 0;
}


void activate_jobs()
{
    int retval;

    /** Wait for the transitioner to create instances
     * of the jobs we just created.
     * Otherwise we'll create too many jobs.
     */
    double now = dtime();
    while (1)
    {
        daemon_sleep(1);
        double x;

        retval = min_transition_time(x);
        if (retval)
        {
            log_messages.printf(MSG_DEBUG, "min_transition_time failed: %s\n", boincerror(retval));
            break;
        }

        if (x > now)
            break;
    }
}

/*
void delete_retry(MysqlRetry * retry)
{
    std::cerr << __LINE__ << " - delete_retry()" << std::endl;

    char buf[1024];

    std::sprintf(buf, "DELETE FROM `%s` WHERE `id`='%llu' LIMIT 1;", MysqlRetry::mysqlTable.c_str(), retry->id);

    std::cerr << buf << std::endl;

    update_mysql(buf);
}


void insert_retry(MysqlRetry * retry, uint64_t shift, uint64_t halfCount)
{
    std::cerr << __LINE__ << " - insert_retry()" << std::endl;

    char buf[1024];

    std::sprintf(buf, "INSERT INTO `%s` (`host_id`, `from`, `count`, `package_id`, `duplicated`, `duplicate`) VALUES ('%d', '%llu', '%llu', '%llu', '%d', '%llu');", MysqlRetry::mysqlTable.c_str(), retry->host_id, retry->from + shift, halfCount, retry->package_id, retry->duplicated, retry->duplicate);

    std::cerr << buf << std::endl;

    update_mysql(buf);
}


MysqlRetry * get_easiest_retry(MysqlHost * host)
{
    std::cerr << __LINE__ << " - get_easiest_retry() for host_id " << host->host_id << std::endl;

    MysqlRetry * retry = NULL;

    char buf[1024];

    std::sprintf(buf, "SELECT * FROM `%s` WHERE package_id = '%d' AND host_id != '%d' ORDER BY count ASC LIMIT 1;", MysqlRetry::mysqlTable.c_str(), host->package_id, host->id);

    std::cerr << buf << std::endl;

    update_mysql(buf);

    MYSQL_RES* rp;

    rp = mysql_store_result(boinc_db.mysql);
    if (!rp)
    {
        std::printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }

    MYSQL_ROW row;

    while ((row = mysql_fetch_row(rp)))
    {
        retry = new MysqlRetry(row);
    }

    mysql_free_result(rp);

    return retry;
}


MysqlJob * get_job_from_retry(MysqlHost * host)
{
    std::cerr << __LINE__ << " - get_job_from_retry() for host_id " << host->host_id << std::endl;

    MysqlRetry * retry = NULL;
    unsigned long long int half1, half2;
    unsigned long long int seconds;

    while(1)
    {
        retry = get_easiest_retry(host);
        if(retry == NULL)
            return NULL;

        seconds = host->calculate_seconds();

        if(retry->count > seconds * host->power)
        {
            half1 = retry->count / 2;
            half2 = retry->count - half1;

            delete_retry(retry);

            insert_retry(retry, 0, half1);
            insert_retry(retry, half1, half2);
        }
        else
        {
            break;
        }
    }

    MysqlJob * job = new MysqlJob(host->id, retry->from, retry->count, false, retry->duplicate);

    delete_retry(retry);

    return job;
}

*/
void clean_error_jobs()
{
    /*
    SELECT id FROM `fc_job` WHERE workunit_id IN (SELECT workunitid from result WHERE `validate_state`=2);
    */
    //TODO
}


/*******************************************/
/** @section MysqlJob function definitions */
/*******************************************/

MysqlJob::MysqlJob(MYSQL_ROW row)
{
    this->m_id = boost::lexical_cast<uint64_t>(row[0]);
    this->m_package_id = boost::lexical_cast<uint64_t>(row[1]);
    this->m_workunit_id = boost::lexical_cast<uint64_t>(row[2]);
    this->m_host_id = boost::lexical_cast<uint64_t>(row[3]);
    this->m_boinc_host_id = boost::lexical_cast<uint64_t>(row[4]);
    this->m_start_index = boost::lexical_cast<uint64_t>(row[5]);
    this->m_hc_keyspace = boost::lexical_cast<uint64_t>(row[6]);
    this->m_duplicated = (row[7][0] == '0') ? true: false;
    this->m_duplicate = boost::lexical_cast<uint64_t>(row[8]);
    this->m_retry = (row[11][0] == '0') ? true: false;
    this->m_finished = (row[12][0] == '0') ? true: false;
}


MysqlJob::MysqlJob(uint64_t package_id, uint64_t host_id, uint64_t boinc_host_id, uint64_t start_index, uint64_t hc_keyspace, bool duplicated, uint64_t duplicate, bool retry)
{
    this->m_id = 0;
    this->m_package_id = package_id;
    this->m_host_id = host_id;
    this->m_boinc_host_id = boinc_host_id;
    this->m_workunit_id = 0;
    this->m_start_index = start_index;
    this->m_hc_keyspace = hc_keyspace;
    this->m_duplicated = duplicated;
    this->m_duplicate = duplicate;
    this->m_retry = retry;
    this->m_finished = false;
}


std::vector<MysqlJob *> MysqlJob::load(std::string filter)
{
    std::vector<MysqlJob *> jobs;
    char buf[SQL_BUF_SIZE];

    std::snprintf(buf, SQL_BUF_SIZE, "SELECT * FROM `%s` %s ;", mysql_table_job.c_str(), filter.c_str());
    update_mysql(buf);

    MYSQL_RES* rp;
    rp = mysql_store_result(boinc_db.mysql);
    if (!rp)
    {
        std::printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }

    MYSQL_ROW row;

    while ((row = mysql_fetch_row(rp)))
    {
        jobs.push_back(new MysqlJob(row));
    }

    mysql_free_result(rp);

    return jobs;
}


unsigned int MysqlJob::row_count(std::string filter)
{
    char buf[SQL_BUF_SIZE];

    std::snprintf(buf, SQL_BUF_SIZE, "SELECT COUNT(*) FROM `%s` %s ;", mysql_table_job.c_str(), filter.c_str());
    update_mysql(buf);

    MYSQL_RES* rp;
    rp = mysql_store_result(boinc_db.mysql);
    if (!rp)
    {
        std::printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }

    MYSQL_ROW row;

    int result = 0;
    while ((row = mysql_fetch_row(rp)))
    {
        result = boost::lexical_cast<unsigned int>(row[0]);
    }

    mysql_free_result(rp);

    return result;
}

/*
MysqlRetry * MysqlJob::duplicate_job(uint64_t package_id_)
{
    std::cerr << __LINE__ << " - MysqlJob::duplicateJob() for package_id " << package_id_ << std::endl;

    char buf[1024];

    std::sprintf(buf, "UPDATE `%s` SET `duplicated`='1' WHERE `id`='%d' LIMIT 1;", mysql_table_job.c_str(), this->id);

    update_mysql(buf);

    MysqlRetry * retry = new MysqlRetry(this->host_id, this->from, this->count, package_id_, 0, this->id);
    return retry;
}
*/

void MysqlJob::add_new_job(uint64_t package_id)
{
    std::cerr << __LINE__ << " - Adding new fc_job entry for package_id = " << package_id << std::endl;

    char buf[SQL_BUF_SIZE];
    std::snprintf(buf, SQL_BUF_SIZE, "INSERT INTO `%s` (`package_id`,`workunit_id`,`host_id`, `boinc_host_id`,`start_index`,`hc_keyspace`,`duplicated`,`duplicate`,`retry`)\
    VALUES ('%" PRIu64 "','%" PRIu64 "','%" PRIu64 "','%" PRIu64 "','%" PRIu64 "','%" PRIu64 "','%d', '%" PRIu64 "', '%d');",
            mysql_table_job.c_str(), this->m_package_id, this->m_workunit_id, this->m_host_id, this->m_boinc_host_id, this->m_start_index, this->m_hc_keyspace, this->m_duplicated, this->m_duplicate, this->m_retry);

    std::cerr << buf << std::endl;

    update_mysql(buf);
}


/********************************************/
/** @section MysqlMask function definitions */
/********************************************/

MysqlMask::MysqlMask(MYSQL_ROW row)
{
    this->m_id = boost::lexical_cast<uint64_t>(row[0]);
    this->m_package_id = boost::lexical_cast<uint64_t>(row[1]);

    if(row[2] == NULL)
        this->m_mask = "";
    else
        this->m_mask = row[2];

    this->m_current_index = boost::lexical_cast<uint64_t>(row[3]);
    this->m_hc_keyspace = boost::lexical_cast<uint64_t>(row[5]);
}


void MysqlMask::update_index(uint64_t current_index)
{
    std::cerr << __LINE__ << " - Updating index of mask " << this->m_mask << " to " << current_index << ". Max keyspace is " << this->m_hc_keyspace << std::endl;

    // Local update
    this->m_current_index = current_index;

    // Databse update
    char buf[SQL_BUF_SIZE];
    std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET current_index = %" PRIu64 " WHERE id = %" PRIu64 " ;", mysql_table_mask.c_str(), current_index, this->m_id);
    update_mysql(buf);
}


std::vector< MysqlMask* > MysqlMask::load(std::string filter)
{
    std::vector<MysqlMask *> masks;

    char buf[SQL_BUF_SIZE];
    std::snprintf(buf, SQL_BUF_SIZE, "SELECT * FROM `%s` %s ;", mysql_table_mask.c_str(), filter.c_str());
    update_mysql(buf);

    MYSQL_RES* rp;
    rp = mysql_store_result(boinc_db.mysql);
    if (!rp)
    {
        std::printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }

    MYSQL_ROW row;

    while ((row = mysql_fetch_row(rp)))
    {
        masks.push_back(new MysqlMask(row));
    }

    mysql_free_result(rp);

    return masks;
}


/***********************************************/
/** @section MysqlPackage function definitions */
/***********************************************/

MysqlPackage::MysqlPackage(MYSQL_ROW row)
{
    this->m_id = boost::lexical_cast<uint64_t>(row[0]);

    if(row[1] == NULL)
        this->m_token = "";
    else
        this->m_token = row[1];

    if(row[2] == NULL)
        this->m_attack = "";
    else
        this->m_attack = row[2];

    this->m_attack_mode = boost::lexical_cast<uint32_t>(row[3]);
    this->m_hash_type = boost::lexical_cast<uint32_t>(row[4]);

    if(row[5] == NULL)
        this->m_hash = "";
    else
        this->m_hash = row[5];

    this->m_status = boost::lexical_cast<uint16_t>(row[6]);

    if(row[7] == NULL)
        this->m_result = "";
    else
        this->m_result = row[7];

    this->m_keyspace = boost::lexical_cast<uint64_t>(row[8]);
    this->m_hc_keyspace = boost::lexical_cast<uint64_t>(row[9]);
    this->m_current_index = boost::lexical_cast<uint64_t>(row[11]);

    if(row[13] == NULL)
        this->m_name = "";
    else
        this->m_name = row[13];

    this->m_seconds_per_job = boost::lexical_cast<uint64_t>(row[18]);

    if(row[19] == NULL)
        this->m_config = "";
    else
        this->m_config = row[19];

    if (row[20] == NULL)
        this->m_dict1 = "";
    else
        this->m_dict1 = row[20];

    if (row[21] == NULL)
        this->m_dict2 = "";
    else
        this->m_dict2 = row[21];

    this->m_replicate_factor = boost::lexical_cast<uint32_t>(row[22]);
}


std::vector<MysqlPackage *> MysqlPackage::load(std::string filter)
{
    std::vector<MysqlPackage *> packages;

    char buf[SQL_BUF_SIZE];

    std::snprintf(buf, SQL_BUF_SIZE, "SELECT * FROM `%s` %s ;", mysql_table_package.c_str(), filter.c_str());

    update_mysql(buf);

    MYSQL_RES* rp;
    rp = mysql_store_result(boinc_db.mysql);
    if (!rp)
    {
        std::printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }

    MYSQL_ROW row;

    while ((row = mysql_fetch_row(rp)))
    {
        packages.push_back(new MysqlPackage(row));
    }

    mysql_free_result(rp);

    return packages;
}


MysqlJob * MysqlPackage::get_oldest_non_duplicated_job(uint64_t host_id)
{
    std::cerr << __LINE__ << " - MysqlPackage::get_oldest_non_duplicated_job() for host_id " << host_id << std::endl;

    MysqlJob * job = NULL;

    char buf[SQL_BUF_SIZE];

    std::snprintf(buf, SQL_BUF_SIZE, "SELECT * FROM `%s` WHERE duplicated = '0' AND host_id != '%" PRIu64 "' AND count > 0 ORDER BY id ASC LIMIT 1;", mysql_table_job.c_str(), host_id);

    std::cerr << buf << std::endl;

    update_mysql(buf);

    MYSQL_RES* rp;
    rp = mysql_store_result(boinc_db.mysql);
    if (!rp)
    {
        std::printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }

    MYSQL_ROW row;

    while ((row = mysql_fetch_row(rp)))
    {
        job = new MysqlJob(row);
    }

    mysql_free_result(rp);

    return job;
}


std::vector<uint64_t> MysqlPackage::find_new_hosts()
{
    std::vector<uint64_t> host_ids;
    char buf[SQL_BUF_SIZE];

    std::snprintf(buf, SQL_BUF_SIZE, "SELECT id FROM host WHERE id IN (SELECT boinc_host_id FROM `%s` WHERE package_id = %" PRIu64 ") AND id NOT IN (SELECT boinc_host_id FROM `%s` WHERE package_id= %" PRIu64 ")",
                  mysql_table_host_activity.c_str(), this->m_id, mysql_table_host.c_str(), this->m_id);
    update_mysql(buf);

    MYSQL_RES* rp;
    rp = mysql_store_result(boinc_db.mysql);
    if (!rp)
    {
        std::printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }

    MYSQL_ROW row;

    uint64_t new_host_id;
    while ((row = mysql_fetch_row(rp)))
    {
        new_host_id = boost::lexical_cast<uint64_t>(row[0]);
        host_ids.push_back(new_host_id);

        std::cerr << __LINE__ << " - Found new host with id = " << new_host_id << std::endl;
    }

    mysql_free_result(rp);
    return host_ids;
}

uint64_t MysqlPackage::get_package_status()
{
    std::cerr << __LINE__ << " - Checking package status for package_id = " << this->m_id << std::endl;

    char buf[SQL_BUF_SIZE];
    std::snprintf(buf, SQL_BUF_SIZE, "SELECT status FROM `%s` WHERE id = %" PRIu64 " LIMIT 1;", mysql_table_package.c_str(), this->m_id);
    uint64_t new_status = get_num_from_mysql(buf);

    std::cerr << __LINE__ << " - Current package status: " << new_status << std::endl;

    this->m_status = new_status;
    return new_status;
}



void MysqlPackage::add_new_hosts()
{
    std::vector<uint64_t> boinc_host_ids;
    std::vector<uint64_t>::iterator boinc_host_id_it;

    boinc_host_ids = find_new_hosts();
    for(boinc_host_id_it = boinc_host_ids.begin(); boinc_host_id_it != boinc_host_ids.end(); ++boinc_host_id_it)
    {
        if (this->get_package_status() >= 10)
            MysqlHost::add_new_host(this->m_id, *boinc_host_id_it);
    }
}


void MysqlPackage::create_benchmarks()
{
    std::vector<MysqlHost *> hosts;
    std::vector<MysqlHost *>::iterator host;
    MysqlJob * job;

    char buf[SQL_BUF_SIZE];

    std::snprintf(buf, SQL_BUF_SIZE, "WHERE status = %d AND time <= now() AND boinc_host_id IN (SELECT boinc_host_id FROM `%s` WHERE package_id = %" PRIu64 ") AND package_id = %" PRIu64 "",
                 Host_benchmark, mysql_table_host_activity.c_str(), this->m_id, this->m_id);

    hosts = MysqlHost::load(buf, this);

    for(host = hosts.begin(); host != hosts.end(); ++host)
    {
        std::snprintf(buf, SQL_BUF_SIZE, "WHERE host_id = %" PRIu64 " AND package_id = %" PRIu64 " AND finished = 0 AND hc_keyspace = 0", (*host)->m_id, this->m_id);
        if(MysqlJob::row_count(buf) == 0)
        {
            std::cerr << __LINE__ << " - Creating new benchmark job for host_id = " << (*host)->m_boinc_host_id << std::endl;
            job = new MysqlJob(this->m_id, (*host)->m_id, (*host)->m_boinc_host_id, 0, 0, false, 0, false);
            make_bench(job, this);
        }
    }
}


void MysqlPackage::update_index(uint64_t current_index)
{
    std::cerr << __LINE__ << " - Updating package current index to: " << current_index << std::endl;

    // Local update
    this->m_current_index = current_index;

    // Database update
    char buf[SQL_BUF_SIZE];
    std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET current_index = %" PRIu64 " WHERE id = %" PRIu64 " ;", mysql_table_package.c_str(), this->m_current_index, this->m_id);
    update_mysql(buf);
}


void MysqlPackage::update_status(uint16_t status)
{
    std::cerr << __LINE__ << " - Updating package to status = " << status << std::endl;

    // Local status
    this->m_status = status;

    // DB status
    char buf[SQL_BUF_SIZE];

    std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET status = %" PRIu16 " WHERE id = %" PRIu64 " LIMIT 1;", mysql_table_package.c_str(), status, this->m_id);
    update_mysql(buf);
}


void MysqlPackage::update_status_running_package(uint16_t status)
{
    std::cerr << __LINE__ << " - Trying to update package " << this->m_id << " to status = " << status << std::endl;

    // Local status
    if (this->get_package_status() >= 10)
        this->m_status = status;

    // DB status
    char buf[SQL_BUF_SIZE];

    std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET status = %" PRIu16 " WHERE id = %" PRIu64 " AND status >= 10 LIMIT 1;", mysql_table_package.c_str(), status, this->m_id);
    update_mysql(buf);

    if (this->m_status == status)
        std::cerr << "Update successful" << std::endl;
    else
        std::cerr << "Update FAILED, status of package is: " << this->m_status << std::endl;
}


void MysqlPackage::create_regular_jobs()
{
    std::vector<MysqlHost *> hosts;
    std::vector<MysqlHost *>::iterator host;
    MysqlJob * job = NULL;

    char buf[SQL_BUF_SIZE];
    std::snprintf(buf, SQL_BUF_SIZE, "WHERE status = %d AND time <= now() AND boinc_host_id IN (SELECT boinc_host_id FROM `%s` WHERE package_id = %" PRIu64 ") AND package_id = %" PRIu64 "",
                 Host_normal, mysql_table_host_activity.c_str(), this->m_id, this->m_id);
    hosts = MysqlHost::load(buf, this);

    /** For each host */
    for(host = hosts.begin(); host != hosts.end(); ++host)
    {
        std::snprintf(buf, SQL_BUF_SIZE, "WHERE host_id = %" PRIu64 " AND package_id = %" PRIu64 " AND finished = 0", (*host)->m_id, this->m_id);
        if(MysqlJob::row_count(buf) >= 2)
        {
            std::cerr << __LINE__ << " - Host with id = " << (*host)->m_boinc_host_id << " has already enough cracking jobs." << std::endl;
            continue;
        }

        /** Update package status from DB */
        this->get_package_status();

        std::cerr << __LINE__ << " - Serving host with id = " << (*host)->m_boinc_host_id << std::endl;

        /** Status: Package running */
        if(this->m_status == Package_running)
        {
            std::cerr << __LINE__ << " - Status: Package is running" << std::endl;
            //job = get_job_from_retry(*host);
            //if(job == NULL)
            //{
            switch(this->m_attack_mode)
            {
                // Dictionary job
                case 0:
                    std::cerr << __LINE__ << " - Generating new 'dictionary' attack job" << std::endl;
                    job = (*host)->generate_job_dict();
                    break;

                // Mask job
                case 3:
                    std::cerr << __LINE__ << " - Generating new 'mask' attack job" << std::endl;
                    job = (*host)->generate_job_mask();
                    break;

                default:
                    std::cerr << __LINE__ << " - Could not recognize attack mode: " << this->m_attack_mode << std::endl;
                    this->update_status_running_package(Package_malformed);
                    return;
            }
            //}

            if(job == NULL || job->m_hc_keyspace == 0)
            {
                std::cerr << __LINE__ << " - Job or its keyspace is NULL, setting package to FINISHING" << std::endl;
                update_status_running_package(Package_finishing);
                return;
            }

            make_job(job, this);
        }

        /** Status: Package finishing */
        else if(this->m_status == Package_finishing)
        {
            std::cerr << __LINE__ << " - Status: Package is finishing" << std::endl;
            /*
            while(1)
            {
                job = get_job_from_retry(*host);
                if(job != NULL)
                {
                    break;
                }

                break; // TODO:

                job = get_oldest_non_duplicated_job((*host)->id);
                if(job == NULL) {
                    break;
                }

                MysqlRetry * retry = job->duplicate_job(this->id);
                insert_retry(retry, 0, job->count);
            }

            if(job != NULL)
            {
                make_job(job, this->id, "", 'n', this->data, this->configFile, this->dictionary);
            }
            else
            {
                std::cerr << __LINE__ << " - MysqlHost->update_status(): Host_done" << std::endl;
                (*host)->update_status(Host_done);
            }
            */
            (*host)->update_status(Host_done);
        }

        /** Status: Package validating
        if(this->m_status == Package_validating)
        {
            std::cerr << __LINE__ << " - Package is validating" << std::endl;
            job = new MysqlJob((*host)->m_parent->m_id, (*host)->m_id, 1, 0, false, 0, false);
            make_job(job, this->m_id, this->m_result, 'v', this->m_data, this->m_config, this->m_dict1, this->m_dict2);
        }*/
    }

    // Check if all jobs are done => exhausted/timeout
    if (this->m_status == Package_finishing)
    {
        std::snprintf(buf, SQL_BUF_SIZE, "WHERE package_id = %" PRIu64 " AND finished = 0 AND hc_keyspace > 0", this->m_id);
        if(MysqlJob::row_count(buf) == 0)
        {
            std::snprintf(buf, SQL_BUF_SIZE, "SELECT COUNT(*) FROM `%s` WHERE id = %" PRIu64 " AND time_end <= now() ;", mysql_table_package.c_str(), this->m_id);
            uint64_t isTimeout = get_num_from_mysql(buf);
            if(isTimeout > 0)
            {
                std::cerr << __LINE__ << " - Timeout reached, all jobs done, package is timeouted" << std::endl;
                this->update_status_running_package(Package_timeout);
            }
            else
            {
                std::cerr << __LINE__ << " - All jobs done, package is exhausted" << std::endl;
                this->update_status_running_package(Package_exhausted);
            }
        }
    }
}


/********************************************/
/** @section MysqlHost function definitions */
/********************************************/

MysqlHost::MysqlHost(MYSQL_ROW row, MysqlPackage * parent)
{
    this->m_id = boost::lexical_cast<uint64_t>(row[0]);
    this->m_boinc_host_id = boost::lexical_cast<uint64_t>(row[1]);
    this->m_power = boost::lexical_cast<uint64_t>(row[2]);
    this->m_package_id = boost::lexical_cast<uint64_t>(row[3]);
    this->m_status = boost::lexical_cast<uint16_t>(row[4]);
    this->m_parent = parent;
}


std::vector<MysqlHost *> MysqlHost::load(std::string filter, MysqlPackage * parent)
{
    std::vector<MysqlHost *> hosts;
    char buf[SQL_BUF_SIZE];

    std::snprintf(buf, SQL_BUF_SIZE, "SELECT * FROM `%s` %s ;", mysql_table_host.c_str(), filter.c_str());
    update_mysql(buf);

    MYSQL_RES* rp;
    rp = mysql_store_result(boinc_db.mysql);
    if (!rp)
    {
        std::printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }

    MYSQL_ROW row;

    while ((row = mysql_fetch_row(rp)))
    {
        hosts.push_back(new MysqlHost(row, parent));
    }

    mysql_free_result(rp);

    return hosts;
}


uint64_t MysqlHost::get_real_host_id(uint64_t host_id)
{
    char buf[SQL_BUF_SIZE];

    std::snprintf(buf, SQL_BUF_SIZE, "SELECT host_id FROM `%s` WHERE id = %" PRIu64 " LIMIT 1 ;", mysql_table_host.c_str(), host_id);
    update_mysql(buf);

    MYSQL_RES* rp;
    rp = mysql_store_result(boinc_db.mysql);
    if (!rp)
    {
        std::printf ("Problem with db\n");
        boinc_db.close();
        exit(1);
    }

    MYSQL_ROW row;

    uint64_t result = 0;
    if ((row = mysql_fetch_row(rp)))
    {
        result = boost::lexical_cast<uint64_t>(row[0]);
    }

    mysql_free_result(rp);
    return result;
}


void MysqlHost::add_new_host(uint64_t package_id, uint64_t boinc_host_id)
{
    std::cerr << __LINE__ << " - Adding new row to fc_host with host_id = " << boinc_host_id << std::endl;

    char buf[SQL_BUF_SIZE];

    std::snprintf(buf, SQL_BUF_SIZE, "INSERT INTO `%s` (boinc_host_id, package_id) VALUES (%" PRIu64 ", %" PRIu64 ");", mysql_table_host.c_str(), boinc_host_id, package_id);
    update_mysql(buf);
}


void MysqlHost::update_status(uint16_t status)
{
    std::cerr << __LINE__ << " - Updating status for host_id = " << this->m_boinc_host_id << ", status = " << status << std::endl;

    // Local status
    this->m_status = status;

    // DB status
    char buf[SQL_BUF_SIZE];
    std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET status = %" PRIu16 " WHERE id = %" PRIu64 " LIMIT 1;", mysql_table_host.c_str(), status, this->m_id);
    update_mysql(buf);
}


uint64_t MysqlHost::calculate_seconds()
{
    std::cerr << __LINE__ << " - MysqlHost::calculate_seconds() - for host_id " << this->m_boinc_host_id << std::endl;

    char buf[SQL_BUF_SIZE];
    uint64_t minimum = 60;

    std::snprintf(buf, SQL_BUF_SIZE, "SELECT TIMESTAMPDIFF(SECOND, time_start, now()) FROM `%s` WHERE id = %" PRIu64 ";", mysql_table_package.c_str(), this->m_parent->m_id);
    std::cerr << buf << std::endl;

    uint64_t maximum = get_num_from_mysql(buf) + minimum;
    if(maximum > this->m_parent->m_seconds_per_job)
    {
        maximum = this->m_parent->m_seconds_per_job;
    }

    std::snprintf(buf, SQL_BUF_SIZE, "SELECT SUM(power) FROM `%s` WHERE package_id = %" PRIu64 " AND TIMESTAMPDIFF(SECOND, time, now()) < %" PRIu64 ";", mysql_table_host.c_str(), this->m_parent->m_id, 2*maximum);
    std::cerr << buf << std::endl;

    uint64_t totalPower = get_num_from_mysql(buf) + 1;
    std::cerr << "Total power: " << totalPower << std::endl;

    uint64_t seconds = ((this->m_parent->m_hc_keyspace - this->m_parent->m_current_index) / totalPower) * 0.1;
    std::cerr << "Seconds: " << seconds << std::endl;

    std::cerr << "( ( " << this->m_parent->m_hc_keyspace << " - " << this->m_parent->m_current_index << " ) / " << totalPower << " ) * 0.1 = " << seconds << std::endl;

    if (seconds < minimum)
        seconds = minimum;
    else if (seconds > maximum)
        seconds = maximum;

    return seconds;
}


MysqlJob * MysqlHost::generate_job_dict()
{
    uint64_t seconds = calculate_seconds();
    uint64_t count = this->m_power * seconds;

    std::cerr << __LINE__ << " - Host could compute #passwords: " << count << std::endl;

    MysqlJob * job = new MysqlJob(this->m_parent->m_id, this->m_id, this->m_boinc_host_id, this->m_parent->m_current_index, count, false, 0, false);
    this->m_parent->update_index(this->m_parent->m_current_index + count);

    return job;
}


MysqlJob * MysqlHost::generate_job_mask()
{
    //uint64_t seconds = calculate_seconds();
    uint64_t count = this->m_power * this->m_parent->m_seconds_per_job;

    std::cerr << __LINE__ << " - Host could compute #passwords: " << count << std::endl;
    std::cerr << __LINE__ << " - There are " << this->m_parent->m_masks.size() << " masks left for this package." << std::endl;

    // Find the following mask
    bool maskFound = false;
    std::vector<MysqlMask *>::iterator maskIt;
    for (maskIt = this->m_parent->m_masks.begin(); maskIt != this->m_parent->m_masks.end();)
    {
        if ((*maskIt)->m_current_index < (*maskIt)->m_hc_keyspace)
        {
            std::cerr << __LINE__ << " - Mask found (" << (*maskIt)->m_mask << "), current index: " << (*maskIt)->m_current_index << ", keyspace: " << (*maskIt)->m_hc_keyspace << std::endl;
            maskFound = true;
            break;
        }

        // Remove the finished mask
        std::cerr << __LINE__ << " - Skipping mask \"" << (*maskIt)->m_mask << "\", it is done" << std::endl;
        maskIt = this->m_parent->m_masks.erase(maskIt);
    }

    if (!maskFound)
    {
        std::cerr << __LINE__ << " - No masks remaining, couldn't create a job" << std::endl;
        return NULL;
    }

    if((*maskIt)->m_current_index + count > (*maskIt)->m_hc_keyspace)
    {
        count = (*maskIt)->m_hc_keyspace - (*maskIt)->m_current_index;
        std::cerr << __LINE__ << " - Adjusting #passwords, mask too small, new #: " << count << std::endl;
    }

    MysqlJob * job = new MysqlJob(this->m_parent->m_id, this->m_id, this->m_boinc_host_id, (*maskIt)->m_current_index, count, false, 0, false);
    (*maskIt)->update_index((*maskIt)->m_current_index + count);
    return job;
}


/**************************/
/** @section Main section */
/**************************/

void main_loop()
{
    std::vector<MysqlPackage *> packages;
    std::vector<MysqlPackage *>::iterator packageIt;
    std::vector<MysqlMask *>::iterator maskIt;

    char buf[SQL_BUF_SIZE];
    bool generated_anything;    // Debug purposes

    while (1)
    {
        generated_anything = false;
        check_stop_daemons();

        /** Remove hosts with finished packages*/
        std::snprintf(buf, SQL_BUF_SIZE, "DELETE FROM `%s` WHERE package_id IN (SELECT id FROM `%s` WHERE (status = %d OR status = %d) ) ;", mysql_table_host.c_str(), mysql_table_package.c_str(), Package_finished, Package_exhausted);
        update_mysql(buf);

        /** Set pacakges with passed time_end to Finishing status */
        packages = MysqlPackage::load("WHERE status = 10 AND time_end <= now() ORDER BY id ASC");
        for(packageIt = packages.begin(); packageIt != packages.end(); ++packageIt)
        {
            (*packageIt)->update_status_running_package(Package_finishing);
        }

        /** Find *RUNNING* packages */
        packages = MysqlPackage::load("WHERE status >= 10 AND ( time_start <= now() OR time_start IS NULL) ORDER BY id ASC");

        for(packageIt = packages.begin(); packageIt != packages.end(); ++packageIt)
        {
            generated_anything = true;
            std::cerr << __LINE__ << " - Trying to create jobs for running package \"" << (*packageIt)->m_name << "\" ... (status: " << (*packageIt)->m_status << ")" << std::endl;

            /** Set start time to *NOW* if it is NULL */
            std::snprintf(buf, SQL_BUF_SIZE, "UPDATE `%s` SET `time_start`=now() WHERE id = %" PRIu64 " AND `time_start` IS NULL;", mysql_table_package.c_str(), (*packageIt)->m_id);
            update_mysql(buf);

            /** In case of mask attack, find all unfinished masks for package */
            std::snprintf(buf, SQL_BUF_SIZE, "WHERE package_id = %" PRIu64 " AND current_index < hc_keyspace ORDER BY id ASC", (*packageIt)->m_id);
            (*packageIt)->m_masks = MysqlMask::load(buf);

            /** Find new hosts, if limit is not reached, who do not yet participate in the package */
            (*packageIt)->add_new_hosts();

            /** Make benchmark jobs for hosts with *BENCHMARK* status */
            (*packageIt)->create_benchmarks();

            /** Make regular jobs for hosts with *NORMAL* status */
            (*packageIt)->create_regular_jobs();

            std::cerr << std::endl;
        }

        activate_jobs();

        daemon_sleep(1);

        clean_error_jobs();

        if (generated_anything)
            std::cerr << " ----------------- END OF MAIN LOOP ITERATION ----------------- \n" << std::endl;
    }
}


void usage(char *name)
{
    std::fprintf(stderr, "This is an example BOINC work generator.\n"
        "This work generator has the following properties\n"
        "(you may need to change some or all of these):\n"
        "  It attempts to maintain a \"cushion\" of 100 unsent job instances.\n"
        "  (your app may not work this way; e.g. you might create work in batches)\n"
        "- Creates work for the application \"example_app\".\n"
        "- Creates a new input file for each job;\n"
        "  the file (and the workunit names) contain a timestamp\n"
        "  and sequence number, so that they're unique.\n\n"
        "Usage: %s [OPTION]...\n\n"
        "Options:\n"
        "  [ --app X                Application name (default: example_app)\n"
        "  [ --in_template_file     Input template (default: example_app_in)\n"
        "  [ --out_template_file    Output template (default: example_app_out)\n"
        "  [ -d X ]                 Sets debug level to X.\n"
        "  [ -h | --help ]          Shows this help text.\n"
        "  [ -v | --version ]       Shows version information.\n", name);
}


/**
 * @brief Main function
 */
int main(int argc, char** argv)
{
    int i, retval;
    char buf[SQL_BUF_SIZE];

    std::cerr << __LINE__ << " - Starting sample_work_generator ..." << std::endl;

    /** Argument parsing */
    for (i = 1; i < argc; ++i)
    {
        /** Set debug level */
        if (is_arg(argv[i], "d"))
        {
            if (!argv[++i])
            {
                /** Missing debug level argument */
                log_messages.printf(MSG_CRITICAL, "%s requires an argument\n\n", argv[--i]);
                usage(argv[0]);
                exit(1);
            }

            int dl = atoi(argv[i]);
            log_messages.set_debug_level(dl);
            if (dl == 4)
            {
                g_print_queries = true;
            }
        }

        /** Set application name */
        else if (!strcmp(argv[i], "--app"))
        {
            if (!argv[++i])
            {
                /** Missing application name argument */
                log_messages.printf(MSG_CRITICAL, "%s requires an argument\n\n", argv[--i]);
                usage(argv[0]);
                exit(1);
            }

            app_name = argv[i];
        }

/*      Set input template
        else if (!strcmp(argv[i], "--in_template_file"))
        {
            in_template_file = argv[i];
        }

        Set output template
        else if (!strcmp(argv[i], "--out_template_file"))
        {
            out_template_file = argv[i];
        }
*/
        /** Show help message */
        else if (is_arg(argv[i], "h") || is_arg(argv[i], "help"))
        {
            usage(argv[0]);
            exit(0);
        }

        /** Show version info */
        else if (is_arg(argv[i], "v") || is_arg(argv[i], "version"))
        {
            std::printf("%s\n", SVN_VERSION);
            exit(0);
        }

        /** Unknown argument */
        else
        {
            log_messages.printf(MSG_CRITICAL, "unknown command line argument: %s\n\n", argv[i]);
            usage(argv[0]);
            exit(1);
        }
    }

    /** Parse config file */
    retval = config.parse_file();
    if (retval)
    {
        log_messages.printf(MSG_CRITICAL, "Can't parse config.xml: %s\n", boincerror(retval));
        exit(1);
    }

    /** Open database */
    retval = boinc_db.open(config.db_name, config.db_host, config.db_user, config.db_passwd);
    if (retval)
    {
        log_messages.printf(MSG_CRITICAL, "can't open db\n");
        exit(1);
    }

    /** Find application */
    std::snprintf(buf, sizeof(buf), "where name='%s'", app_name);
    if (app.lookup(buf))
    {
        log_messages.printf(MSG_CRITICAL, "can't find app %s\n", app_name);
        exit(1);
    }

    /** Find input templates */
    std::snprintf(buf, sizeof(buf), "templates/%s", in_template_file_bench);
    if (read_file_malloc(config.project_path(buf), in_template_bench))
    {
        log_messages.printf(MSG_CRITICAL, "can't read input template %s\n", buf);
        exit(1);
    }

    std::snprintf(buf, sizeof(buf), "templates/%s", in_template_file_mask);
    if (read_file_malloc(config.project_path(buf), in_template_mask))
    {
        log_messages.printf(MSG_CRITICAL, "can't read input template %s\n", buf);
        exit(1);
    }

    std::snprintf(buf, sizeof(buf), "templates/%s", in_template_file_dict);
    if (read_file_malloc(config.project_path(buf), in_template_dict))
    {
        log_messages.printf(MSG_CRITICAL, "can't read input template %s\n", buf);
        exit(1);
    }

    std::snprintf(buf, sizeof(buf), "templates/%s", in_template_file_combinator);
    if (read_file_malloc(config.project_path(buf), in_template_combinator))
    {
        log_messages.printf(MSG_CRITICAL, "can't read input template %s\n", buf);
        exit(1);
    }

    std::snprintf(buf, sizeof(buf), "templates/%s", in_template_file_rule);
    if (read_file_malloc(config.project_path(buf), in_template_rule))
    {
        log_messages.printf(MSG_CRITICAL, "can't read input template %s\n", buf);
        exit(1);
    }

    /** Set starting values */
    start_time = time(0);
    seqno = 0;

    log_messages.printf(MSG_NORMAL, "Starting\n");
    std::cerr << __LINE__ << " - Entering main loop." << std::endl;

    /** Run the main loop */
    main_loop();
}
