/**
 * @file sample_work_generator.h
 * @authors Lukas Zobal
 * @authors Martin Holkovic
 * @date 2.8.2017
 */

/**********************************/
/** @section Includes/definitions */
/**********************************/

#ifndef WORK_GENERATOR_H
#define	WORK_GENERATOR_H

#include <sys/param.h>
#include <unistd.h>
#include <cstdlib>
#include <string>
#include <cstring>
#include <iostream>
#include <cstdio>
#include <algorithm>
#include <vector>
#include <fstream>
#include <stdint.h>
#include <inttypes.h>

#include "backend_lib.h"
#include "boinc_db.h"
#include "error_numbers.h"
#include "filesys.h"
#include "parse.h"
#include "str_replace.h"
#include "str_util.h"
#include "svn_version.h"
#include "util.h"

#include "sched_config.h"
#include "sched_util.h"
#include "sched_msgs.h"


/*******************************/
/** @section 3-hde table names */
/*******************************/

std::string mysql_table_job =           "fc_job";
std::string mysql_table_package =       "fc_package";
std::string mysql_table_host =          "fc_host";
std::string mysql_table_host_activity = "fc_host_activity";
std::string mysql_table_mask =          "fc_mask";
std::string mysql_table_settings =      "fc_settings";


/******************************/
/** @section Global variables */
/******************************/

const char* app_name = "icdf2c";
std::string DICT_DIR = "/home/boincadm/dictionaries/";

const char* in_template_file_bench = "bench_in";
const char* in_template_file_mask = "mask_in";
const char* in_template_file_dict = "dict_in";
const char* in_template_file_combinator = "combinator_in";
const char* in_template_file_rule = "rule_in";
const char* out_template_file = "app_out";

char * in_template_bench;
char * in_template_mask;
char * in_template_dict;
char * in_template_combinator;
char * in_template_rule;
DB_APP app;
int start_time, seqno;


/*******************************/
/** @section Class definitions */
/*******************************/


/**
 * @brief MysqlJob class definition
 */
class MysqlJob
{
    public:
        /** Constructors */
        MysqlJob(MYSQL_ROW row);
        MysqlJob(uint64_t package_id, uint64_t host_id, uint64_t boinc_host_id, uint64_t start_index, uint64_t hc_keyspace, bool duplicated, uint64_t duplicate, bool retry);

        /** Static functions */
        static std::vector<MysqlJob *> load(std::string filter);
        static unsigned int row_count(std::string filter);

        /** Public functions */
        void add_new_job(uint64_t package_id);
        //MysqlRetry* duplicate_job(uint64_t package_id_);

        /** Table Attributes */
        uint64_t m_id;
        uint64_t m_package_id;
        uint64_t m_workunit_id;
        uint64_t m_host_id;
        uint64_t m_boinc_host_id;

        uint64_t  m_start_index;
        uint64_t  m_hc_keyspace;
        bool      m_duplicated;
        uint64_t  m_duplicate;
        bool      m_retry;
        bool      m_finished;
};


/**
 * @brief MysqlMask class definition
 */
class MysqlMask
{
    public:
        /** Constructor */
        MysqlMask(MYSQL_ROW row);

        /** Static functions */
        static std::vector<MysqlMask *> load(std::string filter);

        /**
         * @brief Updates current_index value in database and class instance
         * @param current_index new index value
         */
        void update_index(uint64_t current_index);

        /** Table Attributes */
        uint64_t    m_id;
        uint64_t    m_package_id;
        std::string m_mask;
        uint64_t    m_current_index;
        uint64_t    m_hc_keyspace;
};


/**
 * @brief MysqlPackage class definition
 */
class MysqlPackage
{
    public:
        /** Constructor */
        MysqlPackage(MYSQL_ROW row);

        /** Static functions */
        static std::vector<MysqlPackage *> load(std::string filter);

        /** Public functions */
        std::vector<uint64_t> find_new_hosts();
        void add_to_package(uint64_t package_id);
        void add_new_hosts();
        void create_benchmarks();
        void create_regular_jobs();
        void update_index(uint64_t current_index);
        void update_status(uint16_t status);

        /**
         * @brief Updates the package status, only if it is still running
         * @param status New package status
         */
        void update_status_running_package(uint16_t status);
        MysqlJob * get_oldest_non_duplicated_job(uint64_t host_id);

        /** Table Attributes */
        uint64_t    m_id;
        std::string m_token;
        std::string m_attack;
        uint32_t    m_attack_mode;
        uint32_t    m_hash_type;
        std::string m_hash;
        uint16_t    m_status;
        std::string m_result;
        uint64_t    m_keyspace;
        uint64_t    m_hc_keyspace;
        uint64_t    m_current_index;
        std::string m_name;
        uint64_t    m_seconds_per_job;
        std::string m_config;
        std::string m_dict1;
        std::string m_dict2;
        uint32_t    m_replicate_factor;

        std::vector<MysqlMask *> m_masks;

    private:

        uint64_t get_package_status();
};


/**
 * @brief MysqlHost class definition
 */
class MysqlHost
{
    public:
        /** Constructor */
        MysqlHost(MYSQL_ROW row, MysqlPackage * parent);

        /** Static functions */
        static std::vector<MysqlHost *> load(std::string filter, MysqlPackage * parent);
        static void add_new_host(uint64_t package_id, uint64_t boinc_host_id);
        static uint64_t get_real_host_id(uint64_t host_id);

        /** Public functions */
        uint64_t calculate_seconds();
        MysqlJob * generate_job_dict();
        MysqlJob * generate_job_mask();
        void update_status(uint16_t status);

        /** Table Attributes */
        uint64_t m_id;
        uint64_t m_boinc_host_id;
        uint64_t m_power;
        uint64_t m_package_id;
        uint16_t m_status;

        MysqlPackage * m_parent;
};


/******************************/
/** @section Global functions */
/******************************/

/**
 * @brief Main generator loop
 */
void main_loop();


/**
 * @brief Prints help message with program usage
 * @param name Name of this program
 */
void usage(char *name);


/**
 * @brief Support function for base64 decoding
 * @param base64 String in base64 encoding
 * @return Decoded string
 */
std::string decode(std::string base64);


/**
 * @brief SQL support function, gets number from DB
 * @param query SQL query to perform
 * @return Expects to get one cell table from query, returns the number the cell contains
 */
uint64_t get_num_from_mysql(char * query);


/**
 * @brief SQL support function, update DB with provided query
 * @param query SQL query to perform
 */
void update_mysql(char * query);


/**
 * @brief This function creates input files for a workUnit, which are then sent to host
 * @param job MysqlJob instance of job connected to current workunit
 * @param package_id packageID of current package
 * @return BOINC dependent, 0 is OK
 */
int make_bench(MysqlJob * job, MysqlPackage * package);


/**
 * @brief This function creates input files for a workUnit, which are then sent to host
 * @param job MysqlJob instance of job connected to current workunit
 * @param package_id packageID of current package
 * @return BOINC dependent, 0 is OK
 */
int make_job(MysqlJob * job, MysqlPackage * package);


/**
 * @brief This function creates input files for a workUnit, which are then sent to host
 * @param job MysqlJob instance of job connected to current workunit
 * @param package_id packageID of current package
 * @return BOINC dependent, 0 is OK
 */
int make_job_dict(MysqlJob * job, MysqlPackage * package);


/**
 * @brief This function creates input files for a workUnit, which are then sent to host
 * @param job MysqlJob instance of job connected to current workunit
 * @param package_id packageID of current package
 * @return BOINC dependent, 0 is OK
 */
int make_job_mask(MysqlJob * job, MysqlPackage * package);



/**
 * @brief BOINC sample function, apparently very important and buggy
 */
void activate_jobs();


/**
 * @brief Clear trash after every cycle, currently empty
 */
void clean_error_jobs();


/**
 * @brief Functions to work with retry table, currently not being used
 */
/*
void insert_retry(MysqlRetry * retry, uint64_t shift, uint64_t halfCount);
void delete_retry(MysqlRetry * retry);
MysqlRetry * get_easiest_retry(MysqlHost * host);
MysqlJob * get_job_from_retry(MysqlHost * host);

*/
#endif  /* WORK_GENERATOR_H */
